#include <fstream>
#include "EnvironMap.h"
#include "rgbe.h"
#include "Halton.h"

EnvironMap::EnvironMap(){
  gv = NULL;
  data = NULL;
  Lavg = SbColor(0.0, 0.0, 0.0);
  rot.setValue(SbVec3f(0.0, 1.0, 0.0), 0.0);
//  dim[0] = 1000;
//  dim[1] = 1000;
//  dim[0] = dim[1] = 1500;
//    dim[0] = dim[1] = 1650;
}

EnvironMap::~EnvironMap(){
  free(data);
  delete [] data3;
  
  for (int i=0; i<dimll[0]; i++)
    delete [] pweights[i];
  delete [] pweights;
  delete [] cweights;
  delete [] samples1;
  delete [] samples2;
}

void EnvironMap::init(GlobalVars *gv_){

  // set GlobalVars pointer
  gv = gv_;

  // set environment map rotation
  // this works by rotating the incoming ray backwards
  // when doing a texture lookup
  float angle = -gv->envMapRotation*M_PI/180.0; // convert degrees to radians, reverse
  rot.setValue(SbVec3f(0.0, 1.0, 0.0), angle);

  float *image;
  
  // use single color
  if (strcmp(gv->environMapFile, "-1") == 0){
    dim[0] = dim[1] = 1;
    image = (float *) malloc(sizeof(float)*3 * dim[X]*dim[Y]);
    data = (SbColor*) malloc(sizeof(SbColor) * dim[X]*dim[Y]);
    image[0] = gv->worldBG[0]*100.0; // turn hectowatts into watts
    image[1] = gv->worldBG[1]*100.0;
    image[2] = gv->worldBG[2]*100.0;
  }
  // read HDR image
  else {
  
    // open file
//  char *filename = "grace_probe.float";
//  char *filename = "stpeters_probe.float";
//    char *filename = "seminar.float";
//  char *filename = "/home/u2/users/beason/c/pane2/env/uffizi_probe.float";
    FILE *file = fopen(gv->environMapFile, "rb");
    if (!file){
      printf("Could not open file %s.\n", gv->environMapFile);
      free(data);
      data = NULL;
      exit(1);
    }

    //
    //    0  ...  i (dim[0])
    //    .
    //    .      index(i,j) = dim[0]*j+i;
    //    .
    //    j
    // (dim[1])
    //    

    RGBE_ReadHeader(file, &dim[X], &dim[Y], NULL);

    // allocate space
    image = (float *)malloc(sizeof(float)*3*dim[X]*dim[Y]);
    data = (SbColor*) malloc(sizeof(SbColor)*dim[0]*dim[1]);
  
    RGBE_ReadPixels_RLE(file, image, dim[X], dim[Y]);
    fclose(file);
  }
  
  // copy into memory
  int badVals = 0;
  for (int j=0; j<dim[Y]; j++)
    for (int i=0; i<dim[X]; i++){
      float *pixel = &image[(dim[0]*j+i)*3];

      // check for bad values
      for (int k=0; k<3; k++)
	if (!finite(pixel[k])){
	  pixel[k] = 0.0;
	  badVals++;
	}

      SbColor c(pixel[0], pixel[1], pixel[2]);
      c *= 1.0/100.0;           // convert to hectowatts!!
      data[dim[0]*(dim[Y]-1-j)+i] = c;   
    }

  if (badVals)
    cerr << "Fixed " << badVals << " bad values in environment map "
	 << gv->environMapFile << ".\n";
  delete image;

  // compute pixel weights
  cerr << "Computing environment map pixel weights\n";

  // downsample
  SbColor *data2;
  dim2[0] = MAX2((dim[0]+1)/2, 1);
  dim2[1] = MAX2((dim[1]+1)/2, 1);
  data2 = new SbColor[dim2[0]*dim2[1]];
  for (int j=0; j<dim2[1]; j++)
    for (int i=0; i<dim2[0]; i++){

      int x0, x1, y0, y1;
      x0=i*2;
      x1=x0+1;
      if (x1>=dim[0]) x1 = x0;
      y0=j*2;
      y1=y0+1;
      if (y1>=dim[1]) y1 = y0;

      // box filter
      SbColor c(0.0, 0.0, 0.0);
      c += data[y0*dim[0]+x0];
      c += data[y0*dim[0]+x1];
      c += data[y1*dim[0]+x0];
      c += data[y1*dim[0]+x1];
      data2[j*dim2[0]+i] = c/4.0;
    }
	   
  // temporarily swap downsampled texture
  int dimt[2];
  dimt[0] = dim[0];
  dimt[1] = dim[1];
  SbColor *datat = data;
  data = data2;
  dim[0] = dim2[0];
  dim[1] = dim2[1];

  // create lat long map
  dimll[0] = dim2[0];
  dimll[1] = MAX2(dim2[0]/2, 1);
  data3 = new SbColor[dimll[0]*dimll[1]];
  
  pweights = new float*[dimll[0]];
  for (int i=0; i<dimll[0]; i++)
    pweights[i] = new float[dimll[1]];
  cweights = new float[dimll[0]];

  // precompute sin(theta)
  float *sint = new float[dimll[1]];
  for (int j=0; j<dimll[1]; j++)
    sint[j] = sin((0.5+j)/dimll[1]*M_PI);

  // loop over lat/longs
  SbVec3f dir;
  float rsum = 0.0;
  Lavg = SbColor(0.0, 0.0, 0.0);
  float Lavg_weights = 0.0;
  
  for (int i=0; i<dimll[0]; i++){
    
    float csum = 0.0;
    for (int j=0; j<dimll[1]; j++){

      float theta, phi;
      theta = (0.5+j)/dimll[1]*M_PI;
      phi = (0.5+i)/dimll[0]*2.0*M_PI;
      phi -= M_PI_2;
      if (phi<0.0)
	phi += M_PI*2.0;
      dir.setValue(sin(theta)*cos(phi),
		   cos(theta),
		   sin(theta)*sin(phi));
      SbColor rad = getRadiance(dir);
      float b = brightness(rad);
      pweights[i][j]=b*sint[j];
      csum += pweights[i][j];

      // debugging
      b = clamp(b/4.0f, 0.0f, 1.0f);
      data3[j*dimll[0]+i] = SbColor(b,b,b);

      // average radiance
      Lavg += rad*sint[j];
      Lavg_weights += sint[j];
    }

    cweights[i] = csum;
    rsum += csum;

    // normalize pixel weights
    for (int j=0; j<dimll[1]; j++)
      pweights[i][j] /= csum;
  }

  
  Lavg /= Lavg_weights;

  // reporting
  //cerr << "Lavg = " << Lavg << "\n";
  
  if ((Lavg[0]+Lavg[1]+Lavg[2])>0.00001)
    gv->sceneHasLights = 1;

  // normalize column weights
  for (int i=0; i<dimll[0]; i++)
    cweights[i] /= rsum;    

  // switch back texture
  dim[0]=dimt[0];
  dim[1]=dimt[1];
  data = datat;
  delete [] data2;
  delete [] sint;

  // precompute samples
  
  // stratified sampling
  int sqrt_samples = (int)sqrt((float)gv->totalLightSamples*2);
  samples = sqrt_samples*sqrt_samples;
  samples1 = new double[samples];
  samples2 = new double[samples];

  // hammersley
  for (int i=0; i<samples; i++){
    samples1[i] = (double)i/samples;
    samples2[i] = RI_LP(i);
  }

//   for (int sampx=0; sampx<sqrt_samples; sampx++){
//     for (int sampy=0; sampy<sqrt_samples; sampy++){

//       // jittered
//       samples1[sampy*sqrt_samples+sampx]
// 	= (sampx + r.ctx->mydrand48()) / sqrt_samples;
//       samples2[sampy*sqrt_samples+sampx]
// 	= (sampy + r.ctx->mydrand48()) / sqrt_samples;
//     }
//   }
  
  return;
}


void EnvironMap::writeData3(){
  // spit out image of weights
  ofstream quick;
  quick.open("map.ppm");
  quick << "P3\n";
  quick << dimll[0] << " "
	<< dimll[1] << "\n";
  quick << "255\n";

  // find max value
  float max = 0.0;
  for (int j=0; j<dimll[1]; j++)
    for (int i=0; i<dimll[0]; i++){
      float m;
      m = cweights[i] * pweights[i][j];
      //m = data3[dimll[0]*j+i][1];
      if (m>max)
 	max = m;
    }
  
  for (int j=0; j<dimll[1]; j++)
    for (int i=0; i<dimll[0]; i++){
      int r = (int)(data3[dimll[0]*j+i][0]*255.0);
      int g = (int)(data3[dimll[0]*j+i][1]*255.0);
      int b = (int)(data3[dimll[0]*j+i][2]*255.0);

      float kf = cweights[i] * pweights[i][j] / max * 25.0 * 255.0;
      int k = (int)(kf);
      r = g = b = k; 
      
      quick << clamp(r, 0, 255) << " ";
      quick << clamp(g, 0, 255) << " ";
      quick << clamp(b, 0, 255) << " ";
      quick << "\n";
    }

  quick.close();
  exit(1);
}

SbColor EnvironMap::impSample(Ray &r, SbVec3f &n, SbVec3f &ng, int simple){

  /////////////////////////////////
  // importance sample the lights
  /////////////////////////////////

  SbColor color(0.0, 0.0, 0.0);

  // stratified sampling
  int sqrt_samples = (int)sqrt((float)gv->lightSamples*2);
  
  // use one sample in final gather
  if (simple)
   sqrt_samples = 1;
//   int num_samples = samples;
  
//   if (simple)
//     num_samples = 1;
  
  for (int sampx=0; sampx<sqrt_samples; sampx++){
    for (int sampy=0; sampy<sqrt_samples; sampy++){

      // jittered
      double eps1 = (sampx + r.ctx->mydrand48()) / sqrt_samples;
      double eps2 = (sampy + r.ctx->mydrand48()) / sqrt_samples;
      //double eps1 = drand48(), eps2 = drand48();

//   for (int i=0; i<num_samples; i++){

//     double eps1 = samples1[i];
//     double eps2 = samples2[i];

//     if (simple){
//       eps1 = r.ctx->mydrand48();
//       eps2 = r.ctx->mydrand48();
//     }

      // pick column
      float sum = 0.0;
      int c;
      for (c=0; c<dimll[0]-1; c++){
	if (eps1<(sum+cweights[c]))
	  break;
	sum += cweights[c];
      }
      eps1 = (eps1 - sum) / cweights[c];

      // pick row
      sum = 0.0;
      int p;
      for (p=0; p<dimll[1]-1; p++){
	if (eps2<(sum+pweights[c][p]))
	  break;
	sum += pweights[c][p];
      }
      eps2 = (eps2 - sum) / pweights[c][p];

      // debugging
//       data3[dimll[0]*p+c] = SbColor(1.0, 0.0, 0.0);
//       data3[dimll[0]*p+c+1] = SbColor(1.0, 0.0, 0.0);
//       data3[dimll[0]*(p+1)+c] = SbColor(1.0, 0.0, 0.0);
//       data3[dimll[0]*(p+1)+c+1] = SbColor(1.0, 0.0, 0.0);
      
      float phi = (c+eps1)/dimll[0]*2.0*M_PI;
      phi -= M_PI_2;
      if (phi<0.0)
	phi += M_PI*2.0;
      
      float theta = (p+eps2)/dimll[1]*M_PI;

      SbVec3f l;
      l.setValue(sin(theta)*cos(phi),
		 cos(theta),
		 sin(theta)*sin(phi));
      
      // dont illuminate lights below horizon
      if (ng.dot(l) <= 0.0){
	continue;
      }

      // don't illuminate area that think they are facing away
      float dot1 = n.dot(l);
      if (dot1<=0.0)
	continue;      

      // trace from point to light point
      // i add a small amount to make it not intersect overlapping surfaces
      // i really hate doing this
      // optimize: fix this vv^^
      // fix: could solve with min_t
      const float t_eps = 0.002;
      Ray lr = Ray(r.x+l*t_eps, l, r.ctx);
      if (gv->pane->tritree.intersect_ray(lr, r.tri))
	continue; // hits something (and not background)

      SbVec3f Le = getRadiance(l);
      float pw = cweights[c] * pweights[c][p] * dimll[0] * dimll[1] /
	(2.0 * M_PI * M_PI * sin(theta));
      float value = dot1 / pw;
	  
      // compute color
      // shirley p.120 eq. 15.5
      color += SbColor(r.diffCol[R] * M_1_PI * Le[R] * value,
		       r.diffCol[G] * M_1_PI * Le[G] * value,
		       r.diffCol[B] * M_1_PI * Le[B] * value) ;

    } // end stratified sampling y
  } // end stratified sampling x

  color /= sqrt_samples*sqrt_samples;

//   }
//   color /= num_samples;
      
  // debugging
  //writeData3();
  
  return(color);
}

float EnvironMap::getSample(double eps1, double eps2,
			    SbVec3f &dir, SbColor &col){
  
      // pick column
      float sum = 0.0;
      int c;
      for (c=0; c<dimll[0]-1; c++){
	if (eps1<(sum+cweights[c]))
	  break;
	sum += cweights[c];
      }
      //eps1 = (eps1 - sum) / cweights[c];

      // pick row
      sum = 0.0;
      int p;
      for (p=0; p<dimll[1]-1; p++){
	if (eps2<(sum+pweights[c][p]))
	  break;
	sum += pweights[c][p];
      }
      //eps2 = (eps2 - sum) / pweights[c][p];
      
      float phi = (c+eps1)/dimll[0]*2.0*M_PI;
      phi -= M_PI_2;
      if (phi<0.0)
	phi += M_PI*2.0;
      
      float theta = (p+eps2)/dimll[1]*M_PI;

      dir.setValue(sin(theta)*cos(phi),
		   cos(theta),
		   sin(theta)*sin(phi));

      col = getRadiance(dir);

      // return pdf
      return(cweights[c] * pweights[c][p] * dimll[0] * dimll[1] /
	      (2.0 * M_PI * M_PI * sin(theta)));
}

SbColor EnvironMap::getRadiance(SbVec3f &dir){

  // u,v are [-1,1]
  int i, j;
  float u, v, r;

  // rotate direction
  SbVec3f nDir;
  rot.multVec(dir, nDir);

  // sometimes this happens because of floating point errors
  if (nDir[Z]>1.0)
    nDir[Z]=1.0;
  else if (nDir[Z]<-1.0)
    nDir[Z]=-1.0;
  
  r = 1.0/M_PI*acos(nDir[Z])/sqrt(nDir[X]*nDir[X]+nDir[Y]*nDir[Y]);
  u = nDir[X]*r;
  v = nDir[Y]*r;

  // translate u,v to [0,1]
  u = (u+1.0)/2.0;
  v = (v+1.0)/2.0;

  
//   // translate to pixel coords
//   i = (int)(u*dim[0]);
//   j = (int)(v*dim[1]);

//   return (data[dim[0]*j+i]);
  
  // calculate texture coordinates
  float i_f, j_f;
  i_f = (dim[0]-1)*u;
  j_f = (dim[1]-1)*v;

  // calculate indices
  //int i, j;
  int i1, j1;
  i  = (int)floor(i_f);
  j  = (int)floor(j_f);
  i  = clamp( i, 0, dim[0]-2);
  j  = clamp( j, 0, dim[1]-2);
  i1 = i+1;
  j1 = j+1;
	 
  // calculate texel ratios
  float sp, tp;
  sp = i_f-i;
  tp = j_f-j;
	 
  // perform hermite smoothing
  //sp = 3.0*sp*sp-2.0*sp*sp*sp;
  //tp = 3.0*tp*tp-2.0*tp*tp*tp;

  SbColor retVal =
      (1.0-sp)*(1.0-tp)*data[dim[0]*j +i ] +
            sp*(1.0-tp)*data[dim[0]*j +i1] +
      (1.0-sp)*tp      *data[dim[0]*j1+i ] +
            sp*tp      *data[dim[0]*j1+i1];

//   if (!finite(retVal[R]) ||
//       !finite(retVal[G]) ||
//       !finite(retVal[B]) )
//     cerr << "badval at (i,j) " << i << "," << j << "!\n";
    
  return (retVal);
}
