//WW''W''''''WWW'''WW''''''''''''
//WW  W       WW   WW            
//WW  W  WWW  WW  WWWW  WW   WWW 
//WWWWW W WW  WW   WW  WW W  WW W
//WW  W W WW  WW   WW  WW W  WW W
//WW  W  WWW WWWW  WW   WW   WW W
//                               
//'''''''''''''''''''''''''''''''

#include "Halton.h"
#include <stdlib.h>
#include <time.h>

void Halton::Number(long i, int base){
  double f = inv_base = 1.0/base;
  value = 0.0;
  while (i>0){
    value += f* (double)(i % base);
    i /= base;
    f *= inv_base;
  }
}

void Halton::Next(){
  double r = 1.0 - value - 0.0000000001;
  if (inv_base < r) value += inv_base;
  else {
    double h = inv_base, hh;
    do {
      hh = h;
      h *= inv_base;
    } while (h >= r);
    value += hh + h - 1.0;
  }
}

//WW''''WW''''''WWW'''WW''WW'WW''W''''''WWW'''WW''''''''''''
//WWW  WWW       WW   WW     WW  W       WW   WW            
//W WWW WW W WW  WW  WWWW WW WW  W  WWW  WW  WWWW  WW   WWW 
//W  W  WW W WW  WW   WW  WW WWWWW W WW  WW   WW  WW W  WW W
//W     WW W WW  WW   WW  WW WW  W W WW  WW   WW  WW W  WW W
//W     WW  WWW WWWW  WW  WW WW  W  WWW WWWW  WW   WW   WW W
//                                                          
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

MultiHalton::MultiHalton(){

  // initialize
  // srand48(time(NULL));
  for (int i=0; i<dim; i++)
    h[i].Number(i+3/*lrand48()*/, primes[i]);

  // precompute table
  for (int i=0; i<dim; i++){   // loop over series

    // start new series
    preh[i].Number(i+3, primes[i]);

    // fill table
    for (int j=0; j<predim; j++){  // loop over entry
      preh[i].Next();
      table[i][j] = preh[i];
    }
  }
}

double MultiHalton::next(int d){
  h[d].Next();
  return(h[d]);
}

double MultiHalton::prenext(int d){
  int i = (int)(drand48()*predim);
  return(table[d][i]);
}

const int MultiHalton::primes[50] = { 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229 };


//WW''W''''''WWW''WWW'''''''''''
//WW W        WW   WW           
//WWW    WW   WW   WW   WW  W WW
//WW W  WWWW  WW   WW  WWWW WW  
//WW  W WW    WW   WW  WW   WW  
//WW  W  WWW WWWW WWWW  WWW WW  
//                              
//''''''''''''''''''''''''''''''


//Quasi-Monte Carlo Integration (1995)  (Make Corrections)  (18 citations)
//William J. Morokoff, Russel E. Caflisch
// says Sobol is better for dim>6

// From Efficient Multidimensional Sampling
// by Thomas Kollig and Alexander Keller

double RI_vdC(uint bits, uint r){
  bits = ( bits << 16)
    | (bits >> 16);
  bits = ((bits & 0x00ff00ff) << 8)
    | ((bits & 0xff00ff00) >> 8);
  bits = ((bits & 0x0f0f0f0f) << 4)
    | ((bits & 0xf0f0f0f0) >> 4);
  bits = ((bits & 0x33333333) << 2)
    | ((bits & 0xcccccccc) >> 2);
  bits = ((bits & 0x55555555) << 1)
    | ((bits & 0xaaaaaaaa) >> 1);

  bits ^= r;

  return (double) bits / (double) 0x100000000LL;
}

double RI_S(uint i, uint r){
  for (uint v = 1U<<31; i; i >>= 1, v ^= v>>1)
    if (i & 1)
      r ^= v;

  return (double) r / (double) 0x100000000LL;
}

double RI_LP(uint i, uint r){
   for (uint v = 1U<<31; i; i >>= 1, v |= v>>1)
     if (i & 1)
       r ^= v;
   
   return (double) r / (double) 0x100000000LL;
}
