//WW''''WW'''''''WW''WWWW''WWW'''''''''''''''''''''WWW''''''''''''''''''
//WWW  WWW       WW  WW  W  WW                      WW                  
//W WWW WW  WWW WWWW WW  W  WW  W WW  WWWW     WWW  WW   WWW  WWWW  WWWW
//W  W  WW W WW  WW  WWWW   WW  W WW WWW      WW W  WW  W WW WWW   WWW  
//W     WW W WW  WW  WW     WW  W WW   WWW    WW    WW  W WW   WWW   WWW
//W     WW  WWW  WW  WW    WWWW  WWW WWWW      WWW WWWW  WWW WWWW  WWWW 
//                                                                      
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#ifndef MATPLUSCLASS
#define MATPLUSCLASS

#include <iostream>
#include <Inventor/SbColor.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/fields/SoSFColor.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/fields/SoSFEnum.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoSFVec2f.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoSubNode.h>

using namespace std;

// like a MaterialPlus
class MatPlus : public SoNode {

  SO_NODE_HEADER(MatPlus);

  public:

  enum ReflModel {
    SCHLICK, DATA, LIBRARY, BSSRDF, HIERSUBSURF
  };
  
  // fields
  
  SoSFEnum model; // model (schlick, data, library, ?)

  // Schlick
  // F0_r,g,b for top surface from Material diffuseColor
  // F0_r,g,b fro bottom surface from Material specularColor
  SoSFFloat rough1, rough2;  // roughness  (0=mirror, 1=diffuse (default))
  SoSFFloat sym1, sym2;      // anisotropy
                             // (0=aniisotropic, 1=isotropic (symmetrical, default))
  SoSFFloat R0;              // reflection at normal incidence
                             // Default: 1
                             // -1 means auto-alculated based on index of refractions

  SoSFInt32 layers;            // 1 (default) or 2
  

  //SoSFFloat transparency; // 0 (opaque) - 1 (clear)
  SoSFFloat ior1;     // OUTSIDE index of refraction Default: 1.0 (air)
  SoSFFloat ior2;     // INSIDE index of refraction Default: 1.5 (glass)
  SoSFColor atten;    // attenuation (fraction) at 1m when going through inside
                      // Default (0.98, 0.99, 0.98)
                      // from specularColor ?


  
  SoSFFloat emisEcc;    // Light eccentricity (default 1 = diffuse)
  SoSFFloat minDist;  // minimum distance for intersection (default 0.00001)
  
  static void initClass();

  // constructor
  MatPlus();
  virtual ~MatPlus();

  protected:

  virtual void GLRender(SoGLRenderAction *action);
  virtual void callback(SoCallbackAction *action);
  virtual void doAction(SoAction *action);

  private:


  //SoSFVec2f g, s, d;         // internal (private, automatically calculated)

  
  // Shader node
  
  // for colors
  //SoMaterial *mat;
  
  // Textures
  //const int maxTextures = 2;
  //Texture *text[maxTextures];
  
  // Material conversion
  // 1-shininess -> rough1, rough2
  // transparency:
  //    0: 0->transparency, 1->layers
  // != 0: transparency->transparency, 2->layers, -1->R0
  
  // Glass:
  //
  // diffuseColor 1 1 1   * R0 = ((n2-n1)/(n2+n1))^2 = 0.04 
  // sigma1 0
  // psi1 1
  //
  // ambientColor 1 1 1
  // sigma2 0
  // psi2 1
  //
  // layers 2
  // transparency 1
  // ior1 1.0
  // ior2 1.5
  // atten .98 .99 .98

  // Ideal diffuse
  //
  // diffuseColor 1 1 1   * R0 = 1.0
  // sigma1 1
  // psi1 1
  // layers 1
  // transparency 0

  // Ideal mirror
  //
  // diffuseColor 1 1 1 * R0 = 1
  // sigma1 0
  // psi1 1
  // layers 1
  // transparency 0
  
  // Green plastic
  //
  // diffuseColor 1 1 1  * R0 = .04 ?
  // sigma1 0.02
  // psi1 1
  //
  // ambientColor .1 .8 .1
  // sigma2 1
  // psi2 1
  //
  // layers 2
  // transparency 0

  // Smooth metal ("stainless" alloy)
  // diffuseColor 1 1 1 * R0 = 0.8
  // sigma1 0
  // psi1 1
  // layers 1
  // transparency 0
  
};

#endif /* MATPLUSCLASS */
