//WWWWWW'''''''WW''''''''''''''''WWW'''''''''''''''WWW''''''''''''''''''
//  WW                            WW                WW                  
//  WW    W WW WW  WWW WWW   WWW  WW   WW      WWW  WW   WWW  WWWW  WWWW
//  WW    WW   WW W WW WW W W WW  WW  WWWW    WW W  WW  W WW WWW   WWW  
//  WW    WW   WW W WW WW W W WW  WW  WW      WW    WW  W WW   WWW   WWW
//  WW    WW   WW  WWW WW W  WWW WWWW  WWW     WWW WWWW  WWW WWWW  WWWW 
//                            WW                                        
//''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''''

#include "GlobalVars.h"
#include "Pane.h"
#include "Triangle.h"
#include "Ray.h"
#include "Texture.h"

int Triangle::errorTris = 0;
float Triangle::emissiveArea = 0;

Triangle::Triangle(){
  vert[X].setValue(0.0, 0.0, 0.0);
  vert[Y].setValue(0.0, 0.0, 0.0);
  vert[Z].setValue(0.0, 0.0, 0.0);
  diffuse[0] = SbColor(0.0, 0.0, 0.0);
  diffuse[1] = SbColor(0.0, 0.0, 0.0);
  diffuse[2] = SbColor(0.0, 0.0, 0.0);
  specular = SbColor(0.0, 0.0, 0.0);
  emission = SbColor(0.0, 0.0, 0.0);
  shininess = 0.0;
  transparency = 0.0;
  normal.setValue(0.0, 1.0, 0.0);
  normals[0].setValue(0.0, 1.0, 0.0);
  normals[1].setValue(0.0, 1.0, 0.0);
  normals[2].setValue(0.0, 1.0, 0.0);
  texture = NULL;
  shiny = 0;
  //edge1 = edge2 = SbVec3f(0.0, 0.0, 0.0);
  area = 0.0;
  emissive = 0;
  shown = 0;
  //lastRayId = 0;
  //lastRayHit = 0;
  //lastT = 0.0;
}

Triangle::Triangle(void *data, SoCallbackAction *action,
		   const SoPrimitiveVertex *v0,
		   const SoPrimitiveVertex *v1,
		   const SoPrimitiveVertex *v2){

  //static int idCounter = 0;
  Pane *pane = (Pane *)data;

  // dont render/include invisible objects
  if (action->getDrawStyle() == SoDrawStyle::INVISIBLE){
    area = 0.0;
    return;
  }

  // misc acceleration
  shown = 0;
  //lastRayId = 0;
  //lastT = 0.0;
  //id = idCounter++;

  // translate and store vertices
  SbMatrix matrix = action->getModelMatrix();
  matrix.multVecMatrix(v0->getPoint(), vert[X]);
  matrix.multVecMatrix(v1->getPoint(), vert[Y]);
  matrix.multVecMatrix(v2->getPoint(), vert[Z]);

  // calculate edges
  SbVec3f edge1 = vert[Y]-vert[X];
  SbVec3f edge2 = vert[Z]-vert[X];

  // calculate area
  /*
  SbVec3f edge3 = vert[Y]-vert[Z];
  double sa2 = edge1.dot(edge1);
  double sb2 = edge2.dot(edge2);
  double sc2 = edge3.dot(edge3);
  area = 0.25*sqrt(4.0*sa2*sb2-(sa2+sb2-sc2)*(sa2+sb2-sc2));
  */
  area = 0.5*(edge1.cross(edge2)).length();

  if (area <= 0.0){
    if (errorTris++ == 0)
      cerr << "\n";
    cerr << "\rWarning! " << errorTris
	 << " triangle(s) with 0 area!       ";
    /*
    printVec(vert[X]);
    printVec(vert[Y]);
    printVec(vert[Z]);
    cerr << endl;
    printVec(edge1);
    printVec(edge2);
    cerr << endl;
    */
  }

  // calculate normals
  SbMatrix matInvTrans = matrix.inverse().transpose();
  matInvTrans.multDirMatrix(v0->getNormal(), normals[0]);
  matInvTrans.multDirMatrix(v1->getNormal(), normals[1]);
  matInvTrans.multDirMatrix(v2->getNormal(), normals[2]);
  normals[0].normalize();
  normals[1].normalize();
  normals[2].normalize();

  // set geometric normal
  if (action->getVertexOrdering() == SoShapeHints::CLOCKWISE)
    normal = (vert[2]-vert[0]).cross(vert[1]-vert[0]);
  else
    normal = (vert[1]-vert[0]).cross(vert[2]-vert[0]);
  normal.normalize();

  // don't let there be zero normals
  if (normals[0].dot(normals[0]) < 0.9)
    normals[0] = normal;
  if (normals[1].dot(normals[1]) < 0.9)
    normals[1] = normal;
  if (normals[2].dot(normals[2]) < 0.9)
    normals[2] = normal;

  // dont let n.ng < 0
  float dot[3];
  dot[0] = normals[0].dot(normal);
  dot[1] = normals[1].dot(normal);
  dot[2] = normals[2].dot(normal);
  if (dot[0]<0.0 || dot[1]<0.0 || dot[2]<0.0){
    
    // flip normals
    if (dot[0]<0.0 && dot[1]<0.0 && dot[2]<0.0){
      normals[0] = -normals[0];
      normals[1] = -normals[1];
      normals[2] = -normals[2];
    }
    else {
      // push negative normals up until they're flat
      for (int i=0; i<3; i++)
	if (dot[i]<0.0){
	  normals[i] += (0.0015-dot[i]) * normal;
	  normals[i].normalize();
	}
    }
  }

  // Get material
  SbColor ambient;
  action->getMaterial(ambient, diffuse[0], specular, emission,
		      shininess, transparency,
		      v0->getMaterialIndex());
  action->getMaterial(ambient, diffuse[1], specular, emission,
		      shininess, transparency,
		      v1->getMaterialIndex());
  action->getMaterial(ambient, diffuse[2], specular, emission,
		      shininess, transparency,
		      v2->getMaterialIndex());

  // bad color?
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      if (!finite(diffuse[i][j])){
	//cerr << "BAD COLOR VALUE!\n";
	diffuse[i] = SbColor(1.0, 0.0, 0.0);
      }

  // convert emissive flux density to emissive flux
  if (pane->gv->useEmisFluxDensity){

    // make sure unitsPerMeter is okay
    if (pane->gv->unitsPerMeter<0.0){
      cerr << "\nSettings unitsPerMeter to 1.0\n";
      pane->gv->unitsPerMeter = 1.0;
      pane->gv->metersPerUnit = 1.0;
      pane->gv->metersPerUnit2 = 1.0;
    }
    
    emission *= area;
  }
  
  // calculate center, centEmis, centNorm
//  center = (vert[X] + vert[Y] + vert[Z])/3.0;
//   centNorm.setValue(
//     1./3.*normals[0][X] + 1./3.*normals[1][X] + 1./3.*normals[2][X],
//     1./3.*normals[0][Y] + 1./3.*normals[1][Y] + 1./3.*normals[2][Y],
//     1./3.*normals[0][Z] + 1./3.*normals[1][Z] + 1./3.*normals[2][Z]
//     );
//   if (centNorm.dot(normal)<0.0) // interpolated normal has to face
//     centNorm = -centNorm;       // same as geometric for DL visibility purposes
  emisAvg = (emission[R] + emission[G] + emission[B]) * 1.0/3.0;
  
  // turn off shininess (if supposed to)
  if (pane->gv->allowReflection == 0)
    shininess = 0.0;

  // set boolean variables
  shiny = emissive = 0;
  int non_zero_specular = 0;
  if (shininess)
    shiny = 1;
  if (specular[R] + specular[G] + specular[B] >0.00001)
    non_zero_specular = 1;
  if (emission[R] + emission[G] + emission[B] >0.00001){
      emissive = 1;
      emissiveArea += area;
  }

  shiny = shiny && non_zero_specular; // this avoids problem of shiny but no specColor
  if (shiny)
    pane->gv->sceneHasShininess = 1;
  if (emissive)
    pane->gv->sceneHasLights = 1;
  
  // get texturing info
  int numComps;
  SbVec2s size;
  const unsigned char *img = action->getTextureImage(size, numComps);
  
  if (img){

    // only works for rgb textures
    if (numComps == 3 || numComps == 1){
    /*

      // process rgb triplets into reflectances
      int bytes = size[0]*size[1]*numComps;
      unsigned char *img2 = new unsigned char[bytes];
      if (img2 == NULL){
	cerr << "Error: Could not allocate memory ("
	     << bytes << " bytes) for texture map!\n";
	exit(1);
      }
      cerr << "Correcting " << bytes << " bytes.\n";
      for (int i=0; i<bytes; i++){
	float x = (float)img[i]/255.0;     // convert to float
	float d = -1.45*x+1.5;       // convert to optical density
	float r = pow(10.0f, -d);     // convert to reflectance
	r = clamp(r, 0.0f, 1.0f);
	img2[i] = (unsigned char)(r*255.0);
//	cerr << "texture: " << int(img[i])
//	     << " -> " << int(img2[i]) << "\n";
//	img2[i] = img[i];
      }
    */
//      unsigned char *img2 = Texture::addTextureMap(img, numComps, size);
      
      texture = new Texture(action->getTextureWrapS(),
			 action->getTextureWrapT(),
			 action->getTextureModel(),
			 action->getTextureBlendColor(),
			 numComps, size, img,
			 pane->gv->bumpMapping,
			 pane->gv->bumpScale);

      SbMatrix tmatrix = action->getTextureMatrix();
      SbVec3f t3;
      SbVec4f t4;

      t4 = v0->getTextureCoords();
      t3.setValue(t4[0], t4[1], 0.0);
      tmatrix.multVecMatrix(t3, t3);
      texture->stex[0] = t3[0];
      texture->ttex[0] = t3[1];

      t4 = v1->getTextureCoords();
      t3.setValue(t4[0], t4[1], 0.0);
      tmatrix.multVecMatrix(t3, t3);
      texture->stex[1] = t3[0];
      texture->ttex[1] = t3[1];

      t4 = v2->getTextureCoords();
      t3.setValue(t4[0], t4[1], 0.0);
      tmatrix.multVecMatrix(t3, t3);
      texture->stex[2] = t3[0];
      texture->ttex[2] = t3[1];

      /*
      // debugging
      printVec(texture->stex);
      printVec(texture->ttex);
      cerr << "\n";
      */
/*
      texture->stex[0] = v0->getTextureCoords()[0];
      texture->stex[1] = v1->getTextureCoords()[0];
      texture->stex[2] = v2->getTextureCoords()[0];
      texture->ttex[0] = v0->getTextureCoords()[1];
      texture->ttex[1] = v1->getTextureCoords()[1];
      texture->ttex[2] = v2->getTextureCoords()[1];
*/
    }
    else {
      cerr << "Illegal number of components: " << numComps << endl;
      exit(1);
    }
  }
  else
    texture = NULL;

}


// calculate barycentric coordinates t
void Triangle::calcBaryCoords(const SbVec3f &pnt, SbVec3f &bc) const{
    
  SbVec3f u, v, w; // basis for plane of triangle
  SbVec3f x, y, z; // coordinates of tri verts wrt above basis
  SbVec3f d;       // coordinates of center
  SbVec3f t;       // barycentric coordinates

  SbVec3f edge1 = vert[Y]-vert[X];
  SbVec3f edge2 = vert[Z]-vert[X];
  
  u = edge1;
  u.normalize();

  v = edge2;
  v = v - v.dot(u)*u;
  v.normalize();

  x = SbVec3f(0.0, 0.0, 0.0);
  y = SbVec3f(u.dot(edge1), v.dot(edge1), 0.0);
  z = SbVec3f(u.dot(edge2), v.dot(edge2), 0.0);
  d = SbVec3f(u.dot(pnt-vert[0]), v.dot(pnt-vert[0]), 0.0);

  // meat of the algorithm. taken from a post by David Watson:
  // http://research.microsoft.com/~hollasch/cgindex/math/barycentric.html
  float b0;
  b0    =   (y[X]-x[X]) * (z[Y]-x[Y]) - (z[X]-x[X]) * (y[Y]-x[Y]);
  bc[0] = ( (y[X]-d[X]) * (z[Y]-d[Y]) - (z[X]-d[X]) * (y[Y]-d[Y]) ) / b0;
  bc[1] = ( (z[X]-d[X]) * (x[Y]-d[Y]) - (x[X]-d[X]) * (z[Y]-d[Y]) ) / b0;
  bc[2] = ( (x[X]-d[X]) * (y[Y]-d[Y]) - (y[X]-d[X]) * (x[Y]-d[Y]) ) / b0;
}

// or let user specify them
void Triangle::useBaryCoords(const double a, const double b,
			     const double c, SbVec3f &bc){
  bc[0] = a;
  bc[1] = b;
  bc[2] = c;
}  
  
// calculate interpolated normal
SbVec3f Triangle::getIntNormal(const SbVec3f &bc) const{
  SbVec3f n;
  
  //n = normal;
  //return(n);
  
  n.setValue(
    bc[0]*normals[0][X] + bc[1]*normals[1][X] + bc[2]*normals[2][X],
    bc[0]*normals[0][Y] + bc[1]*normals[1][Y] + bc[2]*normals[2][Y],
    bc[0]*normals[0][Z] + bc[1]*normals[1][Z] + bc[2]*normals[2][Z]
    );
  n.normalize();

  // bump mapping
  if (texture && texture->bumpMap){



/*
    // find gradient of s
    SbVec3f edge0 = vert[1]-vert[0];
    SbVec3f edge1 = vert[2]-vert[0];
    SbVec3f edge1o = 
    float len0 = edge0.normalize();
    float len1 = edge1.normalize();
    float dsd0 = (texture->stex[1]-texture->stex[0])/len0;
    float dsd1 = (texture->stex[2]-texture->stex[0])/len1;
    float dtd0 = (texture->ttex[1]-texture->ttex[0])/len0;
    float dtd1 = (texture->ttex[2]-texture->ttex[0])/len1;
    SbVec3f grads = dsd0*edge0 + dsd1*edge1;
    SbVec3f gradt = dtd0*edge0 + dtd1*edge1;
    grads.normalize();
    gradt.normalize();
    //SbVec3f n2 = grads.cross(gradt);
    //n2.normalize();

    return(grads);
    
    // try edge0
    float ds0;
    SbVec3f dpds;    
    ds0 = texture->stex[1]-texture->stex[0];
    if (ds0 == 0.0){

      // try edge1
      ds0 = texture->stex[2]-texture->stex[0];
      if (ds0 == 0.0)
	dpds = SbVec3f(0.0, 0.0, 0.0);
      else
	dpds = edge1.dot(grads)*grads/ds0;
    }
    else {
      dpds = edge0.dot(grads)*grads/ds0;
    }


    float dt0;
    SbVec3f dpdt;
    dt0 = texture->ttex[1]-texture->ttex[0];
    if (dt0 == 0.0){
      
      // try edge1
      dt0 = texture->ttex[2]-texture->ttex[0];
      if (dt0 == 0.0)
	dpdt = SbVec3f(0.0, 0.0, 0.0);
      else
	dpdt = edge1.dot(gradt)*gradt/dt0;
    }
    else {
      dpdt = edge0.dot(gradt)*gradt/dt0;
    }

    if (ds0 == 0.0 && dt0 == 0.0)
      return(n);


    // s x t = n
    if (ds0 == 0.0){
      grads = gradt.cross(n);
      dpds = 0.1f * grads;
      dbds = 0.0;
    }
    else if (dt0 == 0.0){
      gradt = n.cross(grads);
      dpdt = 0.1f * gradt;
      dbdt = 0.0;
    }
*/
    
     SbVec3f db = texture->getNormal(bc);
//   cerr << "Texture color:" << texture->getColor(bc, diffuse) << "\n";
//   cerr << db << "\n";

    // either of these might be zero
     float dbds = db[0];// / (1.0/texture->size[0]);
     float dbdt = db[1];// / (1.0/texture->size[1]);
//    cerr << "dbds=" << dbds << "\n";
//    cerr << "dbdt=" << dbdt << "\n";

    // partial derivatives from PBR by Pharr & Humpreys
    float ds0 = texture->stex[0]-texture->stex[2];
    float ds1 = texture->stex[1]-texture->stex[2];
    float dt0 = texture->ttex[0]-texture->ttex[2];
    float dt1 = texture->ttex[1]-texture->ttex[2];
    SbVec3f edge0 = vert[0]-vert[2];
    SbVec3f edge1 = vert[1]-vert[2];
    SbVec3f ne0 = normals[0]-normals[2];
    SbVec3f ne1 = normals[1]-normals[2];
    float det = ds0*dt1-dt0*ds1;

    if (det == 0.0)
      return(n);

    SbVec3f dpds = ( dt1*edge0 - dt0*edge1)/det;///texture->size[0];
    SbVec3f dpdt = (-ds1*edge0 + ds0*edge1)/det;///texture->size[1];
    SbVec3f dnds = ( dt1*ne0   - dt0*ne1)  /det;///texture->size[0];
    SbVec3f dndt = (-ds1*ne0   + ds0*ne1)  /det;///texture->size[1];

    SbVec3f dbpds = dpds + texture->bumpScale*dbds*n + db[2]*dnds;
    SbVec3f dbpdt = dpdt + texture->bumpScale*dbdt*n + db[2]*dndt;

//     Ray r2(r.x, dpds, r.ctx);
//     r2.dir = dpds; r2.show(3);
//     r2.dir = dpdt; r2.show(3);
//     r2.dir = dbpds; r2.show(1);
//     r2.dir = dbpdt; r2.show(1);

    
    SbVec3f n2 = dbpds.cross(dbpdt);
//    r2.dir = n2; r2.show(3);

    // flip normal
    if (n2.dot(normal)<0.0)
      n2.negate();

    // rotate normal
    SbRotation sbr;
    sbr.setValue(normal, n);
    sbr.multVec(n2, n2);

    // make non-negative
     if (n2.dot(normal)<0.0)
       n2 += (0.0015-n2.dot(normal)) * normal;

     // normalize
     n2.normalize();
     
    
    return(n2);
    
/*    
    SbVec3f bumpNormal = texture->getNormal(bc);


    SbVec3f u, v, w;
    u = grads;
    w = grads.cross(gradt); //n; //normal;
    w.normalize();
    if (w.dot(normal)<0.0)
      w=-w;
    return(w);
    v = gradt; //w.cross(u);
    //if (v.dot(gradt)<0.0)
    //  v=-v;
    v.normalize();

    if (grads.length() <= 0.0)
      bumpNormal[X] = 0.0;
    if (gradt.length() <= 0.0)
      bumpNormal[Z] = 0.0;
    
    //n = bumpNormal[X]*u + bumpNormal[Y]*w + (-bumpNormal[Z])*v;
    n = bumpNormal[X]*u + bumpNormal[Y]*w + (-bumpNormal[Z])*v;
    n.normalize();

    if (n.dot(normal)<0.0)
      n=-n;

    //if (n.dot(n2)<0.0)
    //  n=-n;

    // push negative normals up until they're flat
//     if (n.dot(normal)<0.0){
//       n += (0.0015-n.dot(normal)) * normal;
//       n.normalize();
//     }
*/    
  }

  
  return(n);
}

// calculate interpolated diffuse color
SbColor Triangle::getIntDiffColor(const SbVec3f &bc) const{
  SbColor color;
    
  if (texture && texture->bumpMap != 2)
    return(texture->getColor(bc, diffuse));

  color.setValue(
    bc[0]*diffuse[0][R] + bc[1]*diffuse[1][R] + bc[2]*diffuse[2][R],
    bc[0]*diffuse[0][G] + bc[1]*diffuse[1][G] + bc[2]*diffuse[2][G],
    bc[0]*diffuse[0][B] + bc[1]*diffuse[1][B] + bc[2]*diffuse[2][B]
    );

  return(color);
}

// calculate interpolated diffuse emission
SbColor Triangle::getIntEmission(const SbVec3f &bc) const{
  SbColor color = emission;
  
  // modulate emission by texture
  if (texture){
    SbColor tcolor = texture->getColor(bc, diffuse);

    color[0] *= tcolor[0];
    color[1] *= tcolor[1];
    color[2] *= tcolor[2];
  }

  return(color);
}

// point debugging

namespace {
  pthread_mutex_t lock_show = PTHREAD_MUTEX_INITIALIZER;
}

void Triangle::show(SbColor c){
  static double incr = 1.0;
  double scale = 10.0;
  const int showVerts = 0;

  pthread_mutex_lock(&lock_show);
  double myincr = incr;
  incr *= 0.999;
  pthread_mutex_unlock(&lock_show);
  
  // dont show triangles more than once
  if (shown==1)
    return;
  shown = 1;
  
  // Tri debugging
  SoMaterial *mat = new SoMaterial();
  mat->emissiveColor.setValue(c);
  SoSeparator *sep;
  SoTransform *trn;
  SoCube *sph = new SoCube();
  sph->width.setValue(myincr*scale);
  sph->depth.setValue(myincr*scale);
  sph->height.setValue(myincr*scale);

  // show vertex normals
  if (1){

    SbVec3f edge1 = vert[Y]-vert[X];
    double l = edge1.length()/4.0;
    (Ray(vert[X], normals[X], NULL)).show(l, SbColor(1.0, 0.0, 0.0));
    (Ray(vert[Y], normals[Y], NULL)).show(l, SbColor(0.0, 1.0, 0.0));
    (Ray(vert[Z], normals[Z], NULL)).show(l, SbColor(0.0, 0.0, 1.0));
  }

  // first vert
  sep = new SoSeparator();
  trn = new SoTransform();
  trn->translation.setValue(vert[X]);
  sep->addChild(trn);
  sep->addChild(mat);
  sep->addChild(sph);
  if (showVerts)
    scnRoot->addChild(sep);

  // second vert
  sep = new SoSeparator();
  trn = new SoTransform();
  trn->translation.setValue(vert[Y]);
  trn->scaleFactor.setValue(SbVec3f(0.9, 0.9, 0.9));
  sep->addChild(trn);
  sep->addChild(mat);
  sep->addChild(sph);
  if (showVerts)
    scnRoot->addChild(sep);

  // third vert
  sep = new SoSeparator();
  trn = new SoTransform();
  trn->translation.setValue(vert[Z]);
  trn->scaleFactor.setValue(SbVec3f(0.8, 0.8, 0.8));
  sep->addChild(trn);
  sep->addChild(mat);
  sep->addChild(sph);
  if (showVerts)
    scnRoot->addChild(sep);

  // triangle
  sep = new SoSeparator();
  SoCoordinate3 *cor = new SoCoordinate3();
  cor->point.set1Value(0, vert[X]);
  cor->point.set1Value(1, vert[Y]);
  cor->point.set1Value(2, vert[Z]);
  SoTriangleStripSet *tset = new SoTriangleStripSet();
  tset->numVertices.set1Value(0, 3);
  sep->addChild(mat);
  sep->addChild(cor);
  //sep->addChild(tset);
  scnRoot->addChild(sep);
  
}


// edges as inset lines representation
void Triangle::show2(){
  Ray r;
  SbVec3f dir;
  float l;

  // dont show triangles more than once
  if (shown==1)
    return;
  shown = 1;

  showPhoton(vert[0], SbColor(1,0,0));
  showPhoton(vert[1], SbColor(0,1,0));
  showPhoton(vert[2], SbColor(0,0,1));
  
  SbVec3f v0 = vert[0]*0.9 + vert[1]*0.05 + vert[2]*0.05;
  SbVec3f v1 = vert[1]*0.9 + vert[2]*0.05 + vert[0]*0.05;
  SbVec3f v2 = vert[2]*0.9 + vert[0]*0.05 + vert[1]*0.05;
      
  r.pos = v0;
  dir = v1-v0;
  l = dir.normalize();
  r.dir = dir;
  r.show2(l, SbColor(1,0,0));

  r.pos = v1;
  dir = v2-v1;
  l = dir.normalize();
  r.dir = dir;
  r.show2(l, SbColor(0,1,0));

  r.pos = v2;
  dir = v0-v2;
  l = dir.normalize();
  r.dir = dir;
  r.show2(l, SbColor(0,0,1));
}

void Triangle::printGlobalStats(){
  cerr << "Triangles with zero area: " << errorTris << "\n"
       << "Total area of emissive triangles: " << emissiveArea << "\n";
}

ostream& operator<< (ostream& os, const Triangle& x){
  os << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
  os << "(" << x.vert[0] << ", " << x.vert[1] << ", " << x.vert[2] << ")";
  //   << x.id << ")";
  // next statement crashes under -O6
  // os << resetiosflags(ios::showpoint | ios::fixed) << setprecision(6);
  return(os);
}
