//WWWWWW'''''''WW''''''''''''''''WWW'''''''''''''''WWW''''''''''''''''''
//  WW                            WW                WW                  
//  WW    W WW WW  WWW WWW   WWW  WW   WW      WWW  WW   WWW  WWWW  WWWW
//  WW    WW   WW W WW WW W W WW  WW  WWWW    WW W  WW  W WW WWW   WWW  
//  WW    WW   WW W WW WW W W WW  WW  WW      WW    WW  W WW   WWW   WWW
//  WW    WW   WW  WWW WW W  WWW WWWW  WWW     WWW WWWW  WWW WWWW  WWWW 
//                            WW                                        
//''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''''

#ifndef TRIANGLECLASS
#define TRIANGLECLASS

#include "Main.h"
class Texture; // forward declaration of Texture class
#include "Ray.h"

class Triangle {
private:
  static int errorTris;
  static float emissiveArea;
    
public:
  // geometry 
  SbVec3f vert[3];
  SbVec3f normals[3];
  SbVec3f normal;           // Not really needed (except for lights)

  // material
  SbColor diffuse[3];
  SbColor specular;         // Could be stored with shader
  SbColor emission;         // Could be stored with shader (only needed for lights)
  // SbShader *shader;

  // candidates for removal
  Texture *texture;         // Texture object is bloated
  float shininess, transparency;  // Could be stored in shader
  unsigned char shiny, emissive;  // Could be stored in shader
  
  char shown;   // Not necessary

  // hit caching
  // should replace with *Hit pointer to save memory
  //SbVec3f edge1, edge2;
  //int lastRayId;
  //char lastRayHit;
  //double lastT;
  //int id;

  // precomputed to accelerate direct lighting calculation
  // or only needed for lights
  float area;
  float emisAvg;       

  
  Triangle();
  Triangle(void *data, SoCallbackAction *action,
		     const SoPrimitiveVertex *v0,
		     const SoPrimitiveVertex *v1,
		     const SoPrimitiveVertex *v2);
  ~Triangle(){}
  void calcBaryCoords(const SbVec3f &pnt, SbVec3f &bc) const;
  void useBaryCoords(const double a, const double b, const double c,
		     SbVec3f &bc);
  SbVec3f getIntNormal(const SbVec3f &bc) const;
//  SbVec3f getIntNormal(const SbVec3f &bc, Ray &r) const;
  SbColor getIntDiffColor(const SbVec3f &bc) const;
  SbColor getIntEmission(const SbVec3f &bc) const;

  /*
  SbColor getIntSpecColor(const SbVec3f &bc);
  SbColor getIntEmission(const SbVec3f &bc);

  // calculate interpolated shininess
  float getIntShininess(const SbVec3f &bc){
    return(bc[0]*shininess[0] + bc[1]*shininess[1] + bc[2]*shininess[2]);
  }

  // calculate interpolated transparency
  float getIntTransparency(const SbVec3f &bc){
    return(bc[0]*transparency[0] + bc[1]*transparency[1] +
	   bc[2]*transparency[2]);
  }
  */
  
  // return geometric normal
  const SbVec3f& getNormal() const{
    return(normal);
  }

  // draw verts (spheres) + tri (polygon)
  void show(SbColor c = SbColor(1.0, 1.0, 1.0));

  // draw as edge lines, slightly inset
  void show2();

  // display statistical information
  void printGlobalStats();   
};

ostream& operator<< (ostream& os, const Triangle& x);

#endif /* TRIANGLECLASS */
