//WWWWWW'''''''WW''''''''''''''''WWW'''''''WW''''WW''''''''WW''''''''''WWW''
//  WW                            WW       WW              WW           WW  
//  WW    W WW WW  WWW WWW   WWW  WW   WW  WW    WW  WWWW WWWW     WWW  WW  
//  WW    WW   WW W WW WW W W WW  WW  WWWW WW    WW WWW    WW     WW W  WW  
//  WW    WW   WW W WW WW W W WW  WW  WW   WW    WW   WWW  WW     WW    WW  
//  WW    WW   WW  WWW WW W  WWW WWWW  WWW WWWWW WW WWWW   WW      WWW WWWW 
//                            WW                                            
//''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''''''''

#include "TriangleList.h"
#include "intersect_ray_tri.h"
#include "Ray.h"

extern int DEBUG_STATE;

const char TriangleList::first_surf[16] = {
  3, 1, 1, 2,
  2, 4, 2, 2,
  2, 1, 4, 2,
  1, 1, 1, 3
};

void TriangleList::deleteContents(){
  vector<Triangle *>::iterator I = trilist.begin();
  for (; I != trilist.end(); I++){
    delete *I;
  }
}

void TriangleList::drawAll(){
  vector<Triangle *>::iterator I = trilist.begin();

  for (; I != trilist.end(); I++){
    (*I)->show();
  }
}

// this function returns 1 if triangle 1 comes before triangle 2
// (if you are going in direction dir), and 0 otherwise.
int TriangleList::first(const SbVec3f &dir, const Triangle *tri1,
			const Triangle *tri2){
  float rdn1 = dir.dot(tri1->normal);
  float rdn2 = dir.dot(tri2->normal);
  int a1 = rdn1<0.0?0:                8;
  int t1 = (tri1->transparency>0.0) * 4;
  int a2 = rdn2<0.0?0:                2;
  int t2 = (tri2->transparency>0.0) * 1;

  // index into decision array
  switch (first_surf[ a1 | t1 | a2 | t2 ]){
  case 1:
    return 1;
  case 2:
    return 0;
  default:
  case 3:
    return tri1<tri2;
  case 4:
    return tri1>tri2;
  }
}


//int TriangleList::intersect_ray(Ray &r, Triangle* &a, double &min_t,
//				int rayId, Triangle *ignoreTri){
int TriangleList::intersect_ray(Ray &r, Triangle *ignoreTri){
  // triangle intersection code
  double t, u, v;
  r.t = 1e8f;
  r.tri = NULL;

  // loop over triangles
  vector<Triangle *>::iterator I = trilist.begin();
  for (; I != trilist.end(); I++){

    Triangle *tri = *I;

    // skip triangle
    if (tri == ignoreTri)
      continue;

    // draw triangle
    //if (DEBUG_STATE)
    //  tri->show2();
      
    if (intersect_ray_tri(r.pos, r.dir, tri->vert[X],
			  tri->vert[Y], tri->vert[Z],
			  &t, &u, &v)){
/*

       if (t>EPSILON_MRT && t<r.t){
          r.t = t;
          r.tri = tri;

	  r.bc[1] = u;
	  r.bc[2] = v;
       }
*/
      
      const float epsilon = 0.0001;
      
      if (t>epsilon){

	const float delta_t = r.t - t;
	if (delta_t > epsilon)                 // common case
	  goto accept;
	else if (delta_t < -epsilon)           // also common
	  continue;
	else if (first(r.dir, tri, r.tri))     // surfaces overlap, arbitration needed
	  goto accept;
	else
	  continue;
	
      }
      else {

	if (t<-epsilon)
	  continue;

	if (ignoreTri != NULL){

	  // if ignoreTri 2nd, reject
	  if (first(r.dir, tri, ignoreTri))
	    continue;
	}

	if (r.tri != NULL){

	  if (first(r.dir, tri, r.tri))
	    goto accept;
	  else
	    continue;
	}
	else
	  goto accept;
	
      }

      accept:
      
      // store triangle distance and pointer
      r.t = t;
      r.tri = tri;

      // store barycentric coords
      r.bc[1] = u;
      r.bc[2] = v;

    }

  }

  if (r.tri){ // Intersection

    // calculate third barycentric coordinate
    r.bc[0] = 1.0-r.bc[1]-r.bc[2];

    // don't let t be negative -- heck, dont let it be zero either (FIXME?)
    if (r.t<0.00001){
      // debugging
      //cerr << "Setting mint from " << r.t << " and pos="
      //   << r.pos[X] << "," << r.pos[Y] << "," << r.pos[Z] << "\n";
      r.t = 0.00001;
    }
      
    
    return(1);
  }
  else // No intersection
    return(0);
}
