//WWWWWW'''''''WW''''''''''''''''WWW'''''''WWWWWW'''''''''''''''''''''''''WWW'
//  WW                            WW         WW                            WW 
//  WW    W WW WW  WWW WWW   WWW  WW   WW    WW    W WW  WW   WW      WWW  WW 
//  WW    WW   WW W WW WW W W WW  WW  WWWW   WW    WW   WWWW WWWW    WW W  WW 
//  WW    WW   WW W WW WW W W WW  WW  WW     WW    WW   WW   WW      WW    WW 
//  WW    WW   WW  WWW WW W  WWW WWWW  WWW   WW    WW    WWW  WWW     WWW WWWW
//                            WW                                              
//''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''''''''''

#ifndef TRIANGLETREECLASS
#define TRIANGLETREECLASS

#include "Main.h"
#include "Triangle.h"
#include "TriangleList.h"
class Ray; // forward declaration of Ray class

#define EPSILON_OS  0.03    // (Octree Size)
#define EPSILON_OI  0.00005 // (Octree Include)
//#define SHOWTRAV 1

// cell
typedef struct node_st {
  struct node_st *child[8];   // 32 bytes
  double mid[3];              // 24 bytes
  TriangleList *tri_list;     // 4 bytes

#ifdef SHOWTRAV  
  // debugging
  SbVec3f min, max;
  int depth;
#endif
} node;

// parameters struct
typedef struct treeParam {
  Ray *r_p;                 // ray
  Triangle *ignoreTri_p;
  //unsigned int rayId;
  unsigned char a;   // used for octree traversal algorithm
} treeParam;

class TriangleTree {
private:
  const static int delta[8][3];
  int num_tris;
  int octreeDebug;
  int maxItems;
  int maxDepth;
  node *root;
  node *compactNodes;
  TriangleList tri_list;
  unsigned int rayId;
  pthread_mutex_t lock;

public:

  // temporary!
  int rays;
    
  double min[3], max[3];
  double max_dist_of_vert;

  // keep bounding sphere of triangles
  SbVec3f center;
  float radius;

  // also keep bounding sphere of specular triangles
  SbVec3f shiny_min, shiny_max, shiny_center;
  float shiny_radius;

  TriangleTree();
  void deleteNode(node *n);
  ~TriangleTree();
  void resetTree();

  void add(Triangle *t){

    // add triangle
    tri_list.add(t);    
    num_tris++;
  }

  int size(){
    return(num_tris);
  }
  
  void printStats(node *ptr = NULL, int *numCubes = NULL, int *numTris = NULL);
  void setParams(int maxD, int maxI, int debug);
  void refine(int simpleDiv = 0);
  double findDividingPlane(node *n, int axis, double vmin[3], double vmax[3]);
  void refineNode(node *n, int depth, int simpleDiv,
		  double vmin[3], double vmax[3]);
  inline int new_node(double &t1, int i1, double &t2,
		      int i2, double &t3, int i3) const;
  int first_node(double &tx0, double &ty0, double &tz0,
		 double &txm, double &tym, double &tzm) const;
  int intersect_ray(Ray &r, Triangle *ignoreTri = NULL) const;
  int proc_subtree(double &tx0, double &ty0, double &tz0,
		   double &tx1, double &ty1, double &tz1,
		   node *n, treeParam &tp) const;
  int ss_intersect_ray(Ray &r);
  void reorderNodes();
};

#endif /* TRIANGLETREECLASS */
