//WWWW'''''''WW''''''''''''''''''WW''''''''''
//WW  W      WW                              
//WW  W  WW  WWW  W WW  WWW  WWW WW WWW   WWW
//WW  W WWWW WW W W WW W WW W WW WW WW W W WW
//WW  W WW   WW W W WW W WW W WW WW WW W W WW
//WWWW   WWW WWW   WWW  WWW  WWW WW WW W  WWW
//                       WW   WW           WW
//'''''''''''''''''''''WWW''WWW''''''''''WWW'

#include "debug.h"
#include <Inventor/nodes/SoPointSet.h>
#include <Inventor/nodes/SoVertexProperty.h>
#include <pthread.h>

namespace {
  pthread_mutex_t lock_showPoint = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_t lock_showCube = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_t lock_showPhoton = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_t lock_print = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_t lock_scnroot = PTHREAD_MUTEX_INITIALIZER;
}


// debugging
void complain(int line){
  cerr << "failure at line " << line << "\n";
}


ostream& operator<< (ostream& os, const SbVec3f& x){
  os << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
  os << "(" << x[X] << ", " << x[Y] << ", " << x[Z] << ")";
  // next statement crashes under -O6
  // os << resetiosflags(ios::showpoint | ios::fixed) << setprecision(6);
  return(os);
}

void showPoint(SbVec3f p, double radius, SbColor c){
  // Ray debugging
  SoSeparator *sep = new SoSeparator();
  static double h=0.0;
  static int count = 0;
  double s=1.0, v=1.0;

  pthread_mutex_lock(&lock_showPoint);
  if (count++>1000000){
    pthread_mutex_unlock(&lock_showPoint);
    return;
  }
  h += 0.03;
  float myh = h;
  pthread_mutex_unlock(&lock_showPoint);

  // Translate base
  SoTransform *trn = new SoTransform();
  //trn->scaleFactor.setValue(scale, scale, scale);
  trn->translation.setValue(p);
  sep->addChild(trn);

  // Give Color
  SoMaterial *mat = new SoMaterial();
  if (c[0]>=0.0){
//    for (int i=0; i<3; i++)
//      c[i] = 1.0-exp(-2.0*c[i]/10.0);
    mat->diffuseColor.setValue(c);
  }
  else
    mat->diffuseColor.setHSVValue(myh, s, v);
  mat->transparency.setValue(0.5);
  sep->addChild(mat);

  // Add sphere
  SoSphere *sphr = new SoSphere();
  sphr->radius.setValue(radius);
  sep->addChild(sphr);

  pthread_mutex_lock(&lock_scnroot);
  scnRoot->addChild(sep);
  pthread_mutex_unlock(&lock_scnroot);
}

void showCube(SbVec3f min, SbVec3f max, SbVec3f c, int clear){

  // add self to scene graph
  static SoSeparator *showCubeSep = NULL;
  pthread_mutex_lock(&lock_showCube);
  if (showCubeSep == NULL){
    // add cube sep to scene
    showCubeSep = new SoSeparator();
    scnRoot->addChild(showCubeSep);

    // make cubes unpickable
    SoPickStyle *style = new SoPickStyle();
    style->style.setValue(SoPickStyle::UNPICKABLE);
    showCubeSep->addChild(style);
  }
  else if (clear){
    showCubeSep->removeAllChildren();

    // make cubes unpickable
    SoPickStyle *style = new SoPickStyle();
    style->style.setValue(SoPickStyle::UNPICKABLE);
    showCubeSep->addChild(style);
    
    pthread_mutex_unlock(&lock_showCube);
    return;
  }
  pthread_mutex_unlock(&lock_showCube);
  
  // Cube
  SoSeparator *cubeSep = new SoSeparator();
  SoMaterial *cubeMat = new SoMaterial();
  cubeMat->diffuseColor.setValue(c);
  cubeSep->addChild(cubeMat);
  
  /*     ____
       /|1  /|5    z
      / |  / |     |
     /__|0=min4    |__y
    2|  / |6=max  /
     | /  | /    x
    3|/___|7

  */
    
  float vertices[8][3] =
  {  { min[0], min[1], min[2]},   // 0
     { min[0], min[1], max[2]},   // 1
     { max[0], min[1], max[2]},   // 2
     { max[0], min[1], min[2]},   // 3
     { min[0], max[1], min[2]},   // 4
     { min[0], max[1], max[2]},   // 5
     { max[0], max[1], max[2]},   // 6
     { max[0], max[1], min[2]}    // 7
  };

  int32_t indices[40] =
  {
    0, 1, 2, 3,  -1,
    2, 6, 7, 3,  -1,
    4, 7, 6, 5,  -1,
    0, 4, 5, 1,  -1,
    1, 5, 6, 2,  -1,
    0, 3, 7, 4,  -1
  };

  // coordinates
  SoCoordinate3 *cubeCoords = new SoCoordinate3();
  cubeCoords->point.setValues(0,8, vertices);
  cubeSep->addChild(cubeCoords);

  // line set
  SoIndexedLineSet *cubeLineSet = new SoIndexedLineSet();
  cubeLineSet->coordIndex.setValues(0, 40, indices);
  cubeSep->addChild( cubeLineSet );

  // add to scene
  pthread_mutex_lock(&lock_showCube);
  showCubeSep->addChild(cubeSep);
  pthread_mutex_unlock(&lock_showCube);

  
}

void showPhoton(SbVec3f p, SbColor c, int clear){

  pthread_mutex_lock(&lock_showPhoton);
  static SoSeparator *showPhotSep = NULL;
  static int num = 0;
  static SoMaterial *mat = NULL;
  static SoVertexProperty *vProp = NULL;

  // figure purposes
  //p[1] = 100.0;
  
  if (showPhotSep == NULL){

    // create sep
    showPhotSep = new SoSeparator();

    // make cubes unpickable
    SoPickStyle *style = new SoPickStyle();
    style->style.setValue(SoPickStyle::UNPICKABLE);
    showPhotSep->addChild(style);

    // add material 
    mat = new SoMaterial();
    showPhotSep->addChild(mat);

    // add draw style
    SoDrawStyle *drawStyle = new SoDrawStyle();
    drawStyle->pointSize.setValue(1.0); //10
    showPhotSep->addChild(drawStyle);
    
    // add PointSet with vertexProperty
    vProp = new SoVertexProperty();
    vProp->materialBinding = SoVertexProperty::PER_VERTEX;
    SoPointSet *pntSet = new SoPointSet();
    pntSet->vertexProperty.setValue(vProp);
    showPhotSep->addChild(pntSet);
    
    // add phot sep to scene
    pthread_mutex_lock(&lock_scnroot);
    scnRoot->addChild(showPhotSep);
    pthread_mutex_unlock(&lock_scnroot);
  }
  else if (clear){
    num = 0;
    vProp->vertex.setNum(0);

    pthread_mutex_unlock(&lock_showPhoton);
    return;
  }

  // add point
  if (num<1000000){
    mat->diffuseColor.set1Value(num,
				clamp(c[0], (float)0.0, (float)1.0),
				clamp(c[1], (float)0.0, (float)1.0),
				clamp(c[2], (float)0.0, (float)1.0));
    vProp->vertex.set1Value(num, p);
    num++;
  }
  
  pthread_mutex_unlock(&lock_showPhoton);
}

void print(char const *fmt, ...){
  va_list ap;
  va_start(ap, fmt);

  pthread_mutex_lock(&lock_print);
  vfprintf(stderr, fmt, ap);
  pthread_mutex_unlock(&lock_print);
  
  va_end(ap);
}


void drawHist(float *flux_map, int n_phi, int n_theta, float scale,
	      SbVec3f x, SbVec3f n){

  SbRotation sbr;
  float phi, theta;
  SbVec3f neww;
  sbr.setValue(SbVec3f(0.0, 1.0, 0.0),
	       n);

  
  // display histogram
  static SoSeparator *histSep = NULL;
  int c_idx=0, f_idx=0;
  SoCoordinate3 *coord;
  SoIndexedFaceSet *faceSet;
  if (!histSep){
    histSep = new SoSeparator();
    coord = new SoCoordinate3();
    faceSet = new SoIndexedFaceSet();
    histSep->addChild(coord);
    histSep->addChild(faceSet);
    scnRoot->addChild(histSep);
  }
  else {
    coord = (SoCoordinate3 *) histSep->getChild(0);
    faceSet = (SoIndexedFaceSet *) histSep->getChild(1);
    faceSet->coordIndex.setNum(0);
    coord->point.setNum(0);
  }

  // add point to coord
  int centIdx = c_idx++;
  coord->point.set1Value(centIdx, x);
    
  for (int u=0; u<n_theta; u++)
    for (int v=0; v<n_phi; v++){

      //   ^     2 3
      // u |     
      // v 0->1  0 1
	
      // 0
      theta = acos(sqrt( u/(double)n_theta ));
      phi = v/(double)n_phi*2.0*M_PI;
      neww.setValue(sin(theta)*cos(phi),
		    cos(theta),
		    sin(theta)*sin(phi));
      sbr.multVec(neww, neww);
      coord->point.set1Value(c_idx+0,x+flux_map[u*n_phi+v]*scale*neww);

      // 1
      theta = acos(sqrt( u/(double)n_theta ));
      phi = (v+1)/(double)n_phi*2.0*M_PI;
      neww.setValue(sin(theta)*cos(phi),
		    cos(theta),
		    sin(theta)*sin(phi));
      sbr.multVec(neww, neww);
      coord->point.set1Value(c_idx+1,x+flux_map[u*n_phi+v]*scale*neww);

      // 2
      theta = acos(sqrt( (u+1)/(double)n_theta ));
      phi = v/(double)n_phi*2.0*M_PI;
      neww.setValue(sin(theta)*cos(phi),
		    cos(theta),
		    sin(theta)*sin(phi));
      sbr.multVec(neww, neww);
      coord->point.set1Value(c_idx+2,x+flux_map[u*n_phi+v]*scale*neww);

      // 3
      theta = acos(sqrt( (u+1)/(double)n_theta ));
      phi = (v+1)/(double)n_phi*2.0*M_PI;
      neww.setValue(sin(theta)*cos(phi),
		    cos(theta),
		    sin(theta)*sin(phi));
      sbr.multVec(neww, neww);
      coord->point.set1Value(c_idx+3,x+flux_map[u*n_phi+v]*scale*neww);


      // draw faces
      faceSet->coordIndex.set1Value(f_idx++, centIdx);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+0);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+1);
      faceSet->coordIndex.set1Value(f_idx++, -1);

      faceSet->coordIndex.set1Value(f_idx++, centIdx);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+1);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+3);
      faceSet->coordIndex.set1Value(f_idx++, -1);

      faceSet->coordIndex.set1Value(f_idx++, centIdx);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+3);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+2);
      faceSet->coordIndex.set1Value(f_idx++, -1);

      faceSet->coordIndex.set1Value(f_idx++, centIdx);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+2);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+0);
      faceSet->coordIndex.set1Value(f_idx++, -1);
	       
      faceSet->coordIndex.set1Value(f_idx++, c_idx+0);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+2);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+3);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+1);
      faceSet->coordIndex.set1Value(f_idx++, -1);
       
      c_idx += 4;
    }
}
