//WW''''WW'''''''''''''''''''''''''''''''''WWWW'''''''WW''''''''''''
//WWW  WWW                                 WW  W      WW            
//W WWW WW  WW  WWWW W   WW   W WW W WW    WW  W  WW  WWW  W WW  WWW
//W  W  WW WWWW WW WW W WW W  WW   W WW    WW  W WWWW WW W W WW W WW
//W     WW WW   WW WW W WW W  WW   W WW    WW  W WW   WW W W WW W WW
//W     WW  WWW WW WW W  WW   WW    WWW    WWWW   WWW WWW   WWW  WWW
//                                   WW                           WW
//'''''''''''''''''''''''''''''''''WWW''''''''''''''''''''''''''WWW'

#define DWORD unsigned int
void AddTrack(DWORD addr,  DWORD asize,  const char *fname, DWORD lnum);
void RemoveTrack(DWORD addr);

inline void *operator new(size_t size, const char *file, int line){
  void *ptr = (void *)malloc(size);
  AddTrack((DWORD)ptr, size, file, line);
  return(ptr);
}

inline void *operator new[](size_t size, const char *file, int line){
  void *ptr = (void *)malloc(size);
  AddTrack((DWORD)ptr, size, file, line);
  return(ptr);
}

inline void operator delete(void *p){
  RemoveTrack((DWORD)p);
  free(p);
}

inline void operator delete[](void *p, size_t size){
  RemoveTrack((DWORD)p);
  free(p);
}

typedef struct {
  DWORD	address;
  DWORD	size;
  char	file[64];
  DWORD	line;
} ALLOC_INFO;

typedef list<ALLOC_INFO*> AllocList;

AllocList *allocList;

void AddTrack(DWORD addr,  DWORD asize,  const char *fname, DWORD lnum){
  ALLOC_INFO *info;

  if(!allocList) {
    allocList = new(AllocList);
  }

  info = new(ALLOC_INFO);
  info->address = addr;
  strncpy(info->file, fname, 63);
  info->line = lnum;
  info->size = asize;
  allocList->insert(allocList->begin(), info);
}

void RemoveTrack(DWORD addr){
  AllocList::iterator i;

  if(!allocList)
    return;
  for(i = allocList->begin(); i != allocList->end(); i++)
    {
      if((*i)->address == addr)
	{
	  allocList->remove((*i));
	  break;
	}
    }
}

void OutputDebugString(char* c){
  cerr << c << flush;
}

void DumpUnfreed(){
  AllocList::iterator i;
  DWORD totalSize = 0;
  char buf[1024];

  if(!allocList)
    return;

  for(i = allocList->begin(); i != allocList->end(); i++) {
    sprintf(buf, "%-25s:\t\tLINE %d,\t\tADDRESS %d\t%d unfreed\n",
	    (*i)->file, (*i)->line, (*i)->address, (*i)->size);
    OutputDebugString(buf);
    totalSize += (*i)->size;
  }
  sprintf(buf, "-----------------------------------------------------------\n");
  OutputDebugString(buf);
  sprintf(buf, "Total Unfreed: %d bytes\n", totalSize);
  OutputDebugString(buf);
}

#define new new(__FILE__, __LINE__)

